#!/usr/bin/env python2 
'''
Copyright (C) 2005 Aaron Spike, aaron@ekips.org

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
'''


'''
Adjustments and changes for the L-system extension by Pieter Barendrecht, 2012

- Added field "Line Width"
- Added field "Colors"

- Added specific syntax for scaling, change in color, change in
  line width, using the symbols @, #, and &
  
- Scaling: @2 scales with a factor 2, @I3 (I stands for Inverse)
  scales with 1/3, @Q5 (Q stands for Square Root) with sqrt(5),
  and @IQ2 with the inverse of the square root of 2.  
  Note: currently only 1 digit for scaling available, so @1.2 is
  not yet possible (just change  your step size for now).
  
- Colors: Add hexadecimal color codes to the field, either by typing 
  them directly or dragging them from Inkscape's color palette. They
  should be separated by ";". Also, there is the option for a semi-
  random color: R(#123456,20) where the 20 (it can be any two-digit
  number) is the random change in lightness/darkness (in percentage)
  with relation to the chosen color.  
  Technical note: hexadecimal colors are in RGB, they are converted
  to HSL, where only the L (lightness) part is affected.  
  To change color in the L-system, use the symbol # either in the
  Axiom or the Rules.  
  
- To change the line width (for now it decreases with 1px every
  iteration), use the symbol & either in the Axiom or the Rules.
  The start value for the line width is the one from the "Line Width"
  field.
'''

'''
Examples:

--- Simple Triangles
Axiom = X
Rules = X=F++F++F++-#@Q2X
Colors = #aa0000;#71c837;#ffcc00;#892ca0;#ff8800
Left Angle = 60
Right Angle = 10 (or something similar)

--- Pythagoras Tree
Axiom: F++F++FX++F
Rules: X=[&#@IQ2-FY--F--FF++F++FX++FF];Y=[&#@IQ2+FY--F--FF++F++FX++FF]
Colors: #502d16;#502d16;#653018;#653018;#784421;#784421;R(#5aa02c,10)
Length: 80
Line With: 8
Angles = 45
Iterations: At least 7 to see the random green colors.

--- Penrose Tiling (quite complicated...)
Axiom = [N]++[N]++[N]++[N]++[N]
Rules = M=OA++PA----NA[-OA----MA]++;N=+OA--PA[---MA--NA]+;O=-MA++NA[+++OA++PA]-;P=--OA++++MA[+PA++++NA]--NA;A=
Angles = 36

This last part, "A=", is very important.

'''

import inkex, simplestyle, pturtle, random, colorsys
from math import sqrt

def stripme(s):
    return s.strip()

class LSystem(inkex.Effect):
    def __init__(self):
        inkex.Effect.__init__(self)
        self.OptionParser.add_option("-o", "--order",
                        action="store", type="int", 
                        dest="order", default=2,
                        help="number of iteration")
        self.OptionParser.add_option("-l", "--langle",
                        action="store", type="float", 
                        dest="langle", default=60.0,
                        help="angle for turning left")
        self.OptionParser.add_option("-r", "--rangle",
                        action="store", type="float", 
                        dest="rangle", default=60.0,
                        help="angle for turning right")
        self.OptionParser.add_option("-s", "--step",
                        action="store", type="float", 
                        dest="step", default=200.0,
                        help="step size")
        self.OptionParser.add_option("-p", "--randomizestep",
                        action="store", type="float", 
                        dest="randomizestep", default=0.0,
                        help="randomize step")
        self.OptionParser.add_option("-z", "--randomizeangle",
                        action="store", type="float", 
                        dest="randomizeangle", default=0.0,
                        help="randomize angle")
        self.OptionParser.add_option("-x", "--axiom",
                        action="store", type="string", 
                        dest="axiom", default="XF++F++F",
                        help="initial state of system")
        self.OptionParser.add_option("-u", "--rules",
                        action="store", type="string", 
                        dest="rules", default="F=F-F++F-F;X=@I3X",
                        help="replacement rules")
        self.OptionParser.add_option("-c", "--colors",
                        action="store", type="string", 
                        dest="colors", default="#000000",
                        help="colors of the L-system")
        self.OptionParser.add_option("-w", "--linewidth",
                        action="store", type="float", 
                        dest="linewidth", default=1.0,
                        help="line width")
        self.OptionParser.add_option("-t", "--tab",
                        action="store", type="string",
                        dest="tab")
        self.stack = []
        self.factor = 1.0
        self.factorchange = False
        self.turtle = pturtle.pTurtle()
  
    # Get data from fields, and write out the complete string with __recurse()
    # Then pass this string to __compose_path() 
    def iterate(self):
        self.rules = dict([map(stripme, i.split("=")) for i in self.options.rules.upper().split(";") if i.count("=")==1])
        string = self.__recurse(self.options.axiom.upper(),0)
        self.__compose_path(string)
    
    def draw_path(self,cindex,width):        
        # Before calling draw_path(), make sure that the path isn't empty.
        # Not necessary, but definitely cleaner.
    
        self.color = self.colors[cindex]
        
        # If the color is random
        if self.color[0] == 'R':
            # Get the Red, Green and Blue components
            Red = int(self.color[3:5],16)
            Green = int(self.color[5:7],16)
            Blue = int(self.color[7:9],16)   
            
            # The random change in color should be two digits long
            # So, 01, 02, to 99. 
            
            Difference = int(self.color[10:12])
            
            # Convert the RGB to HSL
            # Red, Green and Blue should be in the interval [0,1]
            
            Hue, Lightness, Saturation = colorsys.rgb_to_hls(Red/255.0,Green/255.0,Blue/255.0)
       
            # Apply change to the Lightness
            Lightness = min(1.0, abs(Lightness+(random.randint(-Difference,Difference)/100.0))) 
            
            # Convert back
            Red, Green, Blue = colorsys.hls_to_rgb(Hue, Lightness, Saturation)
            
            # Red, Green and Blue are now in between 0 and 1
            # Fix this (i.e. they should be in between 0 and 255)            
            NewRed = str(hex(int(Red*255)))[2:]
            NewGreen = str(hex(int(Green*255)))[2:]
            NewBlue = str(hex(int(Blue*255)))[2:]
            
            self.color = '#' + ('0'+NewRed)[-2:] + ('0'+NewGreen)[-2:] + ('0'+NewBlue)[-2:]          
        
        # Change in s (style): sroke-linecap was "butt", is now "square"
        # You can change this of course in Inkscape by pressing Ctrl + Shift + F
                        
        s = {'stroke-linejoin': 'miter', 'stroke-width': str(width) + 'px', 'stroke-opacity': '1.0', 'fill-opacity': '1.0', 
        'stroke': self.color, 'stroke-linecap': 'square', 'fill': 'none'}
        attribs = {'style':simplestyle.formatStyle(s),'d':self.turtle.getPath()}
        inkex.etree.SubElement(self.current_layer,inkex.addNS('path','svg'),attribs)
    
    def turtle_reset(self):
        pos = self.turtle.getpos()               
        # Clear current path
        self.turtle.clean()
        # Pen up is important (don't draw while moving)
        self.turtle.pu()
        self.turtle.setpos(pos)
        self.turtle.pd()
        
    # Process the string now that it's completely written out
    def __compose_path(self, string):
        self.turtle.pu()
        self.turtle.setpos(self.view_center)
        self.turtle.pd()
        # So the drawing direction is now from right to left, this is in many
        # cases more intuitive.
        self.turtle.setheading(180.0)
        
        inverse = False;
        squareroot = False;
        for c in string:
        
            if self.factorchange == True:
            
                # TODO: Instead of these static combinations of operators (i.e. inverse of square root
                # is possible, but not the other way around) a list of the operators should be kept,
                # and it should be completed in reverse order. Example: IQ2. Save Inverse operator, then
                # save Square Root operator. Then first take the square root, and then the inverse.            
           
                if c in '123456789':
                    # TODO: Add support for multiple digits (and a separator) for the scaling.
                    # At the moment, only 1,2,3,4,5,6,7,8 and 9 are possible (with operators of course)                
                    # To do this, create an own "modus"
                    # LIke while ReadingFactor == True, unless not in '1234567890.', continue creating the factor.
                    if inverse == True:
                        if squareroot == True:
                            self.factor = self.factor * (1 / sqrt( float(c) ))
                        else:
                            self.factor = self.factor * (1 / float(c))
                    elif squareroot == True:
                        self.factor = self.factor * sqrt( float(c) )
                    else:
                        self.factor = self.factor * float(c)
                    self.factorchange = False
                    inverse = False
                    squareroot = False
                elif c == 'I':
                    inverse = True
                elif c == 'Q':
                    squareroot = True
                    
            else:
            
                if c in 'ABCDEF':
                    self.turtle.pd()
                    self.turtle.fd(self.options.step * self.factor * (random.normalvariate(1.0, 0.01 * self.options.randomizestep)))
                elif c in 'GHIJKL':
                    self.turtle.pu()
                    self.turtle.fd(self.options.step * self.factor * (random.normalvariate(1.0,  0.01 * self.options.randomizestep)))
                elif c == '@':
                    self.factorchange = True
                elif c == '#':
                
                    # Not necessary, but this is just cleaner to prevent empty paths.
                    if self.colorindex < len(self.colors)-1 or self.colors[self.colorindex][0] == 'R':
                        # Not necessary to draw when there won't be other colors 
                        # Of course possible, but this would add an additional path.
                        if self.turtle.getPath() != '':
                            self.draw_path(self.colorindex,self.width)
                            self.turtle_reset()
                
                    # Increase index
                    if self.colorindex < len(self.colors)-1:
                        self.colorindex += 1                        
                
                elif c == '&':
                    
                    # Similar to the "#" operator above
                    if self.width > 1.0:
                        if self.turtle.getPath() != '':
                            self.draw_path(self.colorindex,self.width)
                            self.turtle_reset()
                        self.width -= 1.0 
                        
                elif c == '+':
                    self.turtle.lt(self.options.langle * (random.normalvariate(1.0,  0.01 * self.options.randomizeangle)))
                elif c == '-':
                    self.turtle.rt(self.options.rangle * (random.normalvariate(1.0,  0.01 * self.options.randomizeangle)))
                elif c == '|':
                    self.turtle.lt(180)
                elif c == '[':
                    self.stack.append([self.turtle.getpos(), self.turtle.getheading(), self.factor, self.colorindex, self.width])
                elif c == ']':
                    self.turtle.pu()
                    
                    # Check current color index
                    old_colorindex = self.colorindex
                    old_width = self.width
                    
                    pos,heading,self.factor,self.colorindex,self.width = self.stack.pop()
                    self.turtle.setpos(pos)
                    self.turtle.setheading(heading)
                    
                    # Check if color or line width changed, or if the color is random
                    if old_colorindex != self.colorindex or self.colors[old_colorindex][0] == 'R' or old_width != self.width:
                        # And check whether the path is empty (perhaps replace these checks to draw_path() 
                        if self.turtle.getPath() != '':
                            self.draw_path(old_colorindex,old_width)
                            self.turtle_reset()
        
        # Draw the last piece of the path
        self.draw_path(self.colorindex,self.width)

    # Write out the complete string (recursive process)
    def __recurse(self,rule,level):
        level_string = ''
        for c in rule:
            if level < self.options.order:
                try:
                    level_string = level_string + self.__recurse(self.rules[c],level+1)
                except KeyError:
                    level_string = level_string + c
            else:
                level_string = level_string + c 
        return level_string
    
    # After __init__(), this is the place where we end up.
    def effect(self):
        self.width = self.options.linewidth
        self.colors = self.options.colors.split(";")
        self.colorindex = 0        

        self.iterate()

if __name__ == '__main__':
    e = LSystem()
    e.affect()


# vim: expandtab shiftwidth=4 tabstop=8 softtabstop=4 fileencoding=utf-8 textwidth=99

